@@ -5,10 +5,11 @@ License:     BSD-style (see the file LICENSE)
 Maintainer:  Ryan Scott
 Portability: Template Haskell
 
-This module reexports all of the functionality of the other modules in this library.
-It also provides a high-level tutorial on @deriving-compat@'s naming conventions and
-best practices. Typeclass-specific information can be found in their respective
-modules.
+This module reexports all of the functionality of the other modules in this library
+(with the exception of "Data.Deriving.Via", which is only available on GHC 8.2 or
+later). This module also provides a high-level tutorial on @deriving-compat@'s
+naming conventions and best practices. Typeclass-specific information can be found
+in their respective modules.
 -}
 module Data.Deriving (
       -- * Backported changes
@@ -74,6 +75,29 @@ The following changes have been backported:
 
 * In GHC 8.2, deriving 'Show' was changed so that it uses an explicit @showCommaSpace@
   method, instead of repeating the code @showString \", \"@ in several places.
+
+* In GHC 8.4, deriving 'Functor' and 'Traverable' was changed so that it uses 'coerce'
+  for efficiency when the last parameter of the data type is at phantom role.
+
+* In GHC 8.4, the `EmptyDataDeriving` proposal brought forth a slew of changes related
+  to how instances for empty data types (i.e., no constructors) were derived. These
+  changes include:
+
+    * For derived `Eq` and `Ord` instances for empty data types, simply return
+      `True` and `EQ`, respectively, without inspecting the arguments.
+
+    * For derived `Read` instances for empty data types, simply return `pfail`
+      (without `parens`).
+
+    * For derived `Show` instances for empty data types, inspect the argument
+      (instead of `error`ing).
+
+    * For derived `Functor` and `Traversable` instances for empty data
+      types, make `fmap` and `traverse` strict in its argument.
+
+    * For derived `Foldable` instances, do not error on empty data types.
+      Instead, simply return the folded state (for `foldr`) or `mempty` (for
+      `foldMap`), without inspecting the arguments.
 -}
 
 {- $derive
